import asyncio
import json
import threading
import time

import websocket


class WebsocketListener(threading.Thread, websocket.WebSocketApp):
    def __init__(
        self,
        ZWaveMe,
        on_message=None,
        on_error=None,
        on_close=None,
        token=None,
        url=None,
    ):
        self._ZWaveMe = ZWaveMe

        threading.Thread.__init__(self)
        websocket.WebSocketApp.__init__(
            self,
            url,
            header={"Authorization": "Bearer " + token if token else ""},
            on_open=self.on_open,
            on_error=on_error,
            on_message=on_message,
            on_close=on_close,
        )

        self.connected = False
        self.last_update = time.time()

    async def connect(self):
        while not self.connected:
            await asyncio.sleep(0.1)
        return self.connected

    def on_open(self, *args):
        self.connected = True
        self.last_update = time.time()

        self.send(
            json.dumps(
                {
                    "event": "httpEncapsulatedRequest",
                    "responseEvent": "get_devices",
                    "data": {"method": "GET", "url": "/ZAutomation/api/v1/devices"},
                }
            )
        )

    def run_forever(
        self, sockopt=None, sslopt=None, ping_interval=0, ping_timeout=None, **kwargs
    ):
        websocket.WebSocketApp.run_forever(
            self,
            sockopt=sockopt,
            sslopt=sslopt,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            **kwargs
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zwave_me_ws']

package_data = \
{'': ['*']}

install_requires = \
['websocket-client>=1.2.1,<2.0.0']

setup_kwargs = {
    'name': 'zwave-me-ws',
    'version': '0.2.5.3',
    'description': 'Library, implementing websocket connection to ZWave-Me',
    'long_description': 'ZWave-Me-WS is a websocket connection library to a ZWave-Me instance\n**Usage**\n\nTo install this package use:\n\n`pip install zwave-me-ws`\n\nTo use the connector:\n\n```\nfrom zwave_me_ws import ZWaveMe\n\napi = ZWaveMe()\n```\n\n',
    'author': 'Dmitry Vlasov',
    'author_email': 'kerbalspacema@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Z-Wave-Me/zwave-me-ws',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import asyncio
import json
import threading
import time

from .helpers import prepare_devices
from .WebsocketListener import WebsocketListener


class ZWaveMe:
    """Main controller class"""

    def __init__(
        self,
        url,
        token=None,
        on_device_create=None,
        on_device_update=None,
        on_new_device=None,
        platforms=None,
    ):
        self.on_device_create = on_device_create
        self.on_device_update = on_device_update
        self.on_new_device = on_new_device
        self.url = url
        self.token = token
        self.platforms = platforms
        self._ws = None
        self._wshost = None
        self.thread = None
        self.devices = []
        self.uuid = None
        self.is_closed = False

    def start_ws(self):
        """Launch thread."""
        self.thread = threading.Thread(target=self.init_websocket)
        self.thread.daemon = True
        self.thread.start()

    async def get_connection(self):
        """verify connection"""
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, self.start_ws)

        try:
            await asyncio.wait_for(self._ws.connect(), timeout=10.0)
            return True

        except asyncio.TimeoutError:
            await self.close_ws()
            return False

    async def wait_for_info(self):
        while not self.uuid:
            await asyncio.sleep(0.1)
        return self.uuid

    async def close_ws(self):
        loop = asyncio.get_event_loop()
        self.is_closed = True
        blocking_tasks = [
            loop.run_in_executor(None, self.thread.join),
            loop.run_in_executor(None, self._ws.close),
        ]
        await asyncio.wait(blocking_tasks)

    async def get_uuid(self):
        """Get uuid info"""
        loop = asyncio.get_event_loop()
        loop.run_in_executor(None, self.get_info)
        try:
            await asyncio.wait_for(self.wait_for_info(), timeout=5.0)
            return self.uuid
        except asyncio.TimeoutError:
            return

    def send_command(self, device_id, command):
        self._ws.send(
            json.dumps(
                {
                    "event": "httpEncapsulatedRequest",
                    "data": {
                        "method": "GET",
                        "url": "/ZAutomation/api/v1/devices/{}/command/{}".format(
                            device_id, command
                        ),
                    },
                }
            )
        )

    def get_devices(self):
        self._ws.send(
            json.dumps(
                {
                    "event": "httpEncapsulatedRequest",
                    "responseEvent": "get_devices",
                    "data": {"method": "GET", "url": "/ZAutomation/api/v1/devices"},
                }
            )
        )

    def get_device_info(self, device_id):
        self._ws.send(
            json.dumps(
                {
                    "event": "httpEncapsulatedRequest",
                    "responseEvent": "get_device_info",
                    "data": {
                        "method": "GET",
                        "url": "/ZAutomation/api/v1/devices/{}".format(device_id),
                    },
                }
            )
        )

    def get_info(self):
        self._ws.send(
            json.dumps(
                {
                    "event": "httpEncapsulatedRequest",
                    "responseEvent": "get_info",
                    "data": {
                        "method": "GET",
                        "url": "/ZAutomation/api/v1/system/first-access",
                    },
                }
            )
        )

    def init_websocket(self):
        # keep websocket open indefinitely
        while True:
            if self.is_closed:
                return
            self._ws = WebsocketListener(
                ZWaveMe=self,
                on_message=self.on_message,
                on_error=self.on_error,
                on_close=self.on_close,
                token=self.token,
                url=self.url,
            )

            try:
                self._ws.run_forever(ping_interval=5)
            finally:
                self._ws.close()
            time.sleep(5)

    def on_message(self, _, utf):
        if utf:
            dict_data = json.loads(utf)
            if "type" not in dict_data.keys():
                return
            try:
                if dict_data["type"] == "get_devices":
                    if "data" not in dict_data or "body" not in dict_data["data"]:
                        return
                    body = json.loads(dict_data["data"]["body"])
                    if "devices" in body["data"]:
                        self.devices = [
                            device
                            for device in body["data"]["devices"]
                            if device["deviceType"] in self.platforms
                        ]
                        self.on_device_create(prepare_devices(self.devices))
                elif dict_data["type"] == "get_device_info":
                    if "data" not in dict_data or "body" not in dict_data["data"]:
                        return
                    body = json.loads(dict_data["data"]["body"])
                    if "id" in body["data"]:
                        new_device = prepare_devices(
                            [
                                body["data"],
                            ]
                        )[0]
                        self.on_new_device(new_device)
                elif dict_data["type"] == "me.z-wave.devices.level":
                    device = prepare_devices(
                        [
                            dict_data["data"],
                        ]
                    )[0]
                    if device.deviceType == "sensorMultilevel":
                        device.level = str(
                            round(float(dict_data["data"]["metrics"]["level"]), 1)
                        )

                    self.on_device_update(device)

                elif dict_data["type"] == "me.z-wave.namespaces.update":
                    for data in dict_data["data"]:
                        if data["id"] == "devices_all":
                            new_devices = [x["deviceId"] for x in data["params"]]
                            devices_to_install = set(new_devices) - set(
                                [x["id"] for x in self.devices]
                            )
                            for device in devices_to_install:
                                self.get_device_info(device)
                elif dict_data["type"] == "get_info":
                    uuid = json.loads(dict_data["data"]["body"])["data"]["uuid"]
                    if uuid and uuid is not None:
                        self.uuid = uuid
            except Exception as e:
                pass

    def on_error(self, *args, **kwargs):
        error = args[-1]

    def on_close(self, _, *args):
        self._ws.connected = False

    def get_ws(self):
        return self._ws

    def get_wshost(self):
        return self._wshost



from dataclasses import dataclass, field
from typing import Union

FIELDS = [
    "id",
    "deviceType",
    "probeType",
    "locationName",
    "manufacturer",
    "firmware",
    "tags",
]
METRICS_SCALE = ["title", "level", "scaleTitle", "min", "max", "color", "isFailed"]
TYPE_LIGHT = "type-light"


@dataclass
class ZWaveMeData:
    id: str
    deviceType: str
    title: str
    level: Union[str, int, float]
    probeType: str = ""
    scaleTitle: str = ""
    min: str = ""
    max: str = ""
    color: dict = field(default_factory=dict)
    isFailed: bool = False
    locationName: str = ""
    manufacturer: str = ""
    firmware: str = ""
    tags: list[str] = field(default_factory=list)


def prepare_devices(devices: list[dict]) -> list[ZWaveMeData]:
    prepared_devices = []
    for device in devices:
        prepared_device = {
            **{key: device[key] for key in FIELDS if key in device},
            **{
                key: device["metrics"][key]
                for key in METRICS_SCALE
                if key in device["metrics"]
            },
        }
        if prepared_device["probeType"] == "siren":
            prepared_device["deviceType"] = "siren"
        elif TYPE_LIGHT in prepared_device["tags"] and prepared_device[
            "deviceType"
        ] not in ["switchRGB", "switchRGBW"]:
            prepared_device["deviceType"] = "lightMultilevel"
            prepared_device["color"] = {"white": prepared_device["level"]}
            prepared_device["level"] = "on" if prepared_device["level"] else "off"
        elif prepared_device["probeType"] == "motor":
            prepared_device["deviceType"] = "motor"
        elif prepared_device["probeType"] == "fan":
            prepared_device["deviceType"] = "fan"
        prepared_devices.append(prepared_device)
    return [ZWaveMeData(**d) for d in prepared_devices]
